#include <stdio.h>
#include <iostream>
#include <ros/ros.h>
#include <sensor_msgs/fill_image.h>
#include <image_transport/image_transport.h>
#include <camera_info_manager/camera_info_manager.h>
#include "indigo.h"

static void reportError(int error_code) {
  ROS_INFO("Occam API Error: %i", error_code);
  abort();
}

class IndigoSDKNode {
public:
  ros::NodeHandle node;
  sensor_msgs::Image img;

  int dev_index;
  int stitching_enabled;
  int auto_exposure_enabled;
  int auto_gain_enabled;
  int exposure;
  int gain;

  OccamDeviceList* device_list;
  OccamDevice* device;
  OccamImage* image;

  std::string camera_name;
  std::string camera_info_url;

  image_transport::CameraPublisher image_pub;

  boost::shared_ptr<camera_info_manager::CameraInfoManager> cinfo;

  IndigoSDKNode() :
    node("~"),
    dev_index(0),
    stitching_enabled(true),
    auto_exposure_enabled(true),
    auto_gain_enabled(true),
    exposure(240),
    gain(32),
    device_list(0),
    device(0) {
    image_transport::ImageTransport it(node);
    image_pub = it.advertiseCamera("image_raw", 1);

    node.param("dev_index", dev_index, dev_index);
    node.param("stitching", stitching_enabled, stitching_enabled);
    node.param("auto_exposure", auto_exposure_enabled, auto_exposure_enabled);
    node.param("auto_gain", auto_gain_enabled, auto_gain_enabled);
    node.param("exposure", exposure, exposure);
    node.param("gain", gain, gain);
    
    ROS_INFO("dev_index: %i",dev_index);
    ROS_INFO("stitching: %i",stitching_enabled);
    ROS_INFO("auto_exposure: %i",auto_exposure_enabled);
    ROS_INFO("auto_gain: %i",auto_gain_enabled);
    ROS_INFO("exposure: %i",exposure);
    ROS_INFO("gain: %i",gain);

    node.param("camera_frame_id", img.header.frame_id, std::string("head_camera"));
    node.param("camera_name", camera_name, std::string("head_camera"));
    node.param("camera_info_url", camera_info_url, std::string(""));
    cinfo.reset(new camera_info_manager::CameraInfoManager(node, camera_name, camera_info_url));

    int r;
    if ((r = occamEnumerateDeviceList(2000, &device_list)) != OCCAM_API_SUCCESS)
      reportError(r);
    ROS_INFO("%i devices found", device_list->entry_count);
    for (int i=0;i<device_list->entry_count;++i) {
      ROS_INFO("device[%i]: cid = %s",i,device_list->entries[i].cid);
    }
    if (dev_index<0 || dev_index >= device_list->entry_count) {
      ROS_INFO("device index %i out of range",dev_index);
      return;
    }

    if ((r = occamOpenDevice(device_list->entries[dev_index].cid, &device)) != OCCAM_API_SUCCESS)
      reportError(r);

    occamSetDeviceValuei(device, OCCAM_STITCHING_ENABLED, stitching_enabled);
    occamSetDeviceValuei(device, OCCAM_AUTO_EXPOSURE, auto_exposure_enabled);
    occamSetDeviceValuei(device, OCCAM_AUTO_GAIN, auto_gain_enabled);
    occamSetDeviceValuei(device, OCCAM_EXPOSURE, exposure);
    occamSetDeviceValuei(device, OCCAM_GAIN, gain);
  }

  virtual ~IndigoSDKNode() {
    if (device)
      occamCloseDevice(device);
    if (device_list)
      occamFreeDeviceList(device_list);
  }

  IndigoSDKNode(const IndigoSDKNode& x) = delete;
  IndigoSDKNode& operator= (const IndigoSDKNode& rhs) = delete;

  bool take_and_send_image() {
    int r;
    if ((r = occamDeviceReadImage(device, &image, 1)) != OCCAM_API_SUCCESS)
      reportError(r);

    const char* image_encoding = 0;
    switch (image->format) {
    case OCCAM_GRAY8: image_encoding = "mono8"; break;
    case OCCAM_RGB24: image_encoding = "rgb8"; break;
    }
    fillImage(img,
	      image_encoding,
	      image->height-1,
	      image->width,
	      image->step[0],
	      image->data[0]+image->step[0]);
    img.header.stamp = ros::Time::now();
    sensor_msgs::CameraInfoPtr ci(new sensor_msgs::CameraInfo(cinfo->getCameraInfo()));
    ci->header.frame_id = img.header.frame_id;
    ci->header.stamp = img.header.stamp;
    image_pub.publish(img, *ci);

    //    ROS_INFO("width %i height %i step %i data %p encoding %s",
    //    image->width,image->height,image->step[0],image->data[0],image_encoding);

    occamFreeImage(image);

    return true;
  }

  bool spin() {
    if (!device)
      return false;

    while (node.ok()) {
      if (!take_and_send_image()) {
        ROS_ERROR("couldn't take image.");
        usleep(1000000);
      }
    }
    return true;
  }
};

int main(int argc, char **argv) {
  int r;
  if ((r = occamInitialize()) != OCCAM_API_SUCCESS)
    reportError(r);
  ros::init(argc, argv, "ros_indigosdk");
  IndigoSDKNode a;
  a.spin();
  exit(0);
  return 0;
}
