#include "device_iface.h"
#include <sstream>
#include <assert.h>
#include <string.h>

////////////////////////////////////////////////////////////////////////
// RegisterProgramError

RegisterProgramError::RegisterProgramError(const std::string& _msg, int _error)
  : msg(_msg),
    error(_error) {
}

RegisterProgramError::~RegisterProgramError() throw() {
}

const char* RegisterProgramError::what() const throw() {
  return msg.c_str();
}

int RegisterProgramError::errorCode() const {
  return error;
}

////////////////////////////////////////////////////////////////////////
// RegisterProgramResponse

void RegisterProgramResponse::add(uint32_t reg_index, uint32_t reg_value){ 
  ResultInfo& res = *results.insert(results.end(),ResultInfo());
  res.reg_index = reg_index;
  res.reg_value = reg_value;
}

void RegisterProgramResponse::assign(int i, uint32_t reg_index, uint32_t reg_value) {
  assert(i>=0&&i<results.size());
  ResultInfo& res = results[i];
  res.reg_index = reg_index;
  res.reg_value = reg_value;
}

void RegisterProgramResponse::clear() {
  results.clear();
}

int RegisterProgramResponse::size() const {
  return results.size();
}

uint32_t RegisterProgramResponse::regIndex(int i) const {
  return results[i].reg_index;
}

uint32_t RegisterProgramResponse::regValue(int i) const {
  return results[i].reg_value;
}

uint32_t RegisterProgramResponse::operator[] (int i) const {
  return results[i].reg_value;
}

std::string RegisterProgramResponse::toString() const {
  std::stringstream sout;
  for (int j=0;j<results.size();++j) {
    const ResultInfo& r = results[j];
    sout<<(j?", ":"")<<std::hex<<"0x"<<r.reg_index<<"=0x"<<r.reg_value<<std::dec;
  }
  return sout.str();
}

////////////////////////////////////////////////////////////////////////
// RegisterProgram

template <class T>
const char* RegisterProgram::parseNumber(const char* cmd_s, T& number, bool* late_bound) {
  int number_type = 0;
  int digit_count = 0;
  number = 0;
  int state = 0;
  const char* chars[] = 
    {"0123456789",
     "0123456789abcdef",
     "01"
    };
  int base[] = {10,16,2};
  if (late_bound)
    *late_bound = false;
  for (;;) {
    if (!*cmd_s) {
      std::stringstream sout;
      sout<<"Syntax error, expected number";
      throw RegisterProgramError(sout.str());
    }
    else if (*cmd_s == ' ' || *cmd_s == '\t' || *cmd_s == '\r' || *cmd_s == '\n')
      ++cmd_s;
    else {
      break;
    }
  }
  for (bool done=false;!done;) {
    switch (state) {
    case 0: {
      assert(*cmd_s);
      if (*cmd_s == '?' && late_bound) {
	++cmd_s;
	*late_bound = true;
	done = true;
	break;
      } else if (*cmd_s == '0') {
	++cmd_s;
	state = 1;
      } else if (*cmd_s == 'b') {
	number_type = 2;
	++cmd_s;
	state = 2;
      } else
	state = 2;
      break;
    }
    case 1: {
      if (*cmd_s != 'x') {
	++digit_count;
	state = 2;
	break;
      }
      number_type = 1;
      ++cmd_s;
      state = 2;
      break;
    }
    case 2: {
      bool valid_char = false;
      int digit_value = 0;
      for (digit_value=0;chars[number_type][digit_value];++digit_value)
	if (*cmd_s == chars[number_type][digit_value]) {
	  valid_char = true;
	  break;
	}
      if (!valid_char && digit_count == 0) {
	std::stringstream sout;
	sout<<"Syntax error, expected at least one digit: "<<chars[number_type];
	throw RegisterProgramError(sout.str());
      }
      if (!valid_char) {
	done = true;
	break;
      }
      number = number * base[number_type];
      number += digit_value;
      ++digit_count;
      ++cmd_s;
      break;
    }
    }
  }
  return cmd_s;
}

template const char* RegisterProgram::parseNumber<uint32_t>(const char* cmd_s, uint32_t& number, bool* late_bound);
template const char* RegisterProgram::parseNumber<uint64_t>(const char* cmd_s, uint64_t& number, bool* late_bound);

const char* RegisterProgram::parseNextChar(const char* cmd_s, char& c, const char* expect_hint) {
  for (;;) {
    if (!*cmd_s) {
      std::stringstream sout;
      sout<<"Syntax error, expected character: "<<expect_hint<<", got end of data";
      throw RegisterProgramError(sout.str());
    }
    else if (*cmd_s == ' ' || *cmd_s == '\t' || *cmd_s == '\r' || *cmd_s == '\n')
      ++cmd_s;
    else {
      c = *cmd_s;
      ++cmd_s;
      break;
    }
  }
  return cmd_s;
}

const char* RegisterProgram::parseDelim(const char* cmd_s, char delim, bool optional) {
  for (;;) {
    if (!*cmd_s && !optional) {
      std::stringstream sout;
      sout<<"Syntax error, expected delimeter: "<<delim;
      throw RegisterProgramError(sout.str());
    }
    else if (!*cmd_s)
      break;
    if (*cmd_s == ' ' || *cmd_s == '\t' || *cmd_s == '\r' || *cmd_s == '\n')
      ++cmd_s;
    else if (*cmd_s == delim) {
      ++cmd_s;
      break;
    }
    else if (*cmd_s != delim) {
      std::stringstream sout;
      sout<<"Syntax error, expected delimeter: "<<delim<<", but got: "<<*cmd_s;
      throw RegisterProgramError(sout.str());
    }
  }
  return cmd_s;
}

const char* RegisterProgram::parseCommandType(const char* cmd_s, CommandType& type) {
  char c;
  cmd_s = parseNextChar(cmd_s, c, "r w W s");
  if (c == 'r')
    type = CommandType_Read;
  else if (c == 'w')
    type = CommandType_Write;
  else if (c == 'W')
    type = CommandType_ReadWrite;
  else if (c == 's')
    type = CommandType_Sleep;
  else {
    std::stringstream sout;
    sout<<"Unknown command type: "<<c;
    throw RegisterProgramError(sout.str());
  }
  return cmd_s;
}

const char* RegisterProgram::parseCommand(const char* cmd_s, CommandInfo& cmd) {
  cmd_s = parseCommandType(cmd_s, cmd.type);
  cmd.late_bound = false;
  if (cmd.type == CommandType_Read) {
    cmd_s = parseNumber(cmd_s, cmd.reg_index);
  } else if (cmd.type == CommandType_Write) {
    cmd_s = parseNumber(cmd_s, cmd.reg_index);
    cmd_s = parseDelim(cmd_s,'=',false);
    cmd_s = parseNumber(cmd_s, cmd.reg_value, &cmd.late_bound);
  } else if (cmd.type == CommandType_ReadWrite) {
    cmd_s = parseNumber(cmd_s, cmd.reg_index);
    cmd_s = parseDelim(cmd_s,'(',false);
    cmd_s = parseNumber(cmd_s, cmd.reg_mask);
    cmd_s = parseDelim(cmd_s,')',false);
    cmd_s = parseDelim(cmd_s,'=',false);
    cmd_s = parseNumber(cmd_s, cmd.reg_value, &cmd.late_bound);
  } else if (cmd.type == CommandType_Sleep) {
    cmd_s = parseNumber(cmd_s, cmd.delay_us);
  }

  cmd_s = parseDelim(cmd_s,';',true);
  return cmd_s;
}

RegisterProgram::RegisterProgram() {
}

RegisterProgram::RegisterProgram(const std::string& reg_cmd) {
  add(reg_cmd);
}

void RegisterProgram::add(const std::string& reg_cmd) {
  if (reg_cmd.empty())
    return;
  const char* cmd_s = reg_cmd.c_str();
  for (;;) {
    for (int j=0;
	 *cmd_s==' '&&*cmd_s=='\t'&&
	   *cmd_s=='r'&&*cmd_s=='\n';++cmd_s);
    if (!*cmd_s)
      break;
    CommandInfo cmd;
    cmd_s = parseCommand(cmd_s, cmd);
    if (!cmd_s)
      break;
    commands.push_back(cmd);
  }
}

void RegisterProgram::addRead(uint32_t reg_index) {
  CommandInfo& cmd = *commands.insert(commands.end(),CommandInfo());
  cmd.type = CommandType_Read;
  cmd.reg_index = reg_index;
}

void RegisterProgram::addWrite(uint32_t reg_index, uint32_t reg_value) {
  CommandInfo& cmd = *commands.insert(commands.end(),CommandInfo());
  cmd.type = CommandType_Write;
  cmd.reg_index = reg_index;
  cmd.reg_value = reg_value;
}

void RegisterProgram::addSleep(uint64_t delay_us) {
  CommandInfo& cmd = *commands.insert(commands.end(),CommandInfo());
  cmd.type = CommandType_Sleep;
  cmd.delay_us = delay_us;
}

void RegisterProgram::bind(int index, uint32_t value) {
  for (CommandInfo& cmd : commands) {
    if (!cmd.late_bound)
      continue;
    if (!index) {
      cmd.reg_value = value;
      break;
    }
    --index;
  }
}

RegisterProgramResponse RegisterProgram::execute(OccamDeviceBase& dev) {
  RegisterProgramResponse resp;
  for (const CommandInfo& cmd : commands) {
    int r = OCCAM_API_SUCCESS;
    switch (cmd.type) {
    case CommandType_Read: {
      uint32_t read_value = 0;
      r = dev.readRegister(cmd.reg_index, &read_value);
      resp.add(cmd.reg_index, read_value);
    }
    case CommandType_Write: {
      r = dev.writeRegister(cmd.reg_index, cmd.reg_value);
      break;
    }
    case CommandType_ReadWrite: {
      r = dev.writeRegister(cmd.reg_index, cmd.reg_value, cmd.reg_mask);
      break;
    }
    case CommandType_Sleep: {
      usleep(cmd.delay_us);
      break;
    }
    }

    if (r != OCCAM_API_SUCCESS)
      throw RegisterProgramError("Register program error", r);
  }
  return resp;
}

int RegisterProgram::size() const {
  return commands.size();
}

const RegisterProgram::CommandInfo& RegisterProgram::operator[] (int index) const {
  return commands[index];
}

RegisterProgram::CommandInfo& RegisterProgram::operator[] (int index) {
  return commands[index];
}

//////////////////////////////////////////////////////////////////////////////////
// OccamDeviceBase

OccamDeviceBase::OccamDeviceBase(const std::string& cid)
  : _cid(cid) {
  std::string::size_type p0 = _cid.find_first_of(":");
  if (p0 != std::string::npos) {
    _model.assign(_cid.begin(),_cid.begin()+p0);
    _serial.assign(_cid.begin()+p0+1,_cid.end());
  }
}

OccamDeviceBase::~OccamDeviceBase() {
}

const std::string& OccamDeviceBase::cid() const {
  return _cid;
}

const std::string& OccamDeviceBase::model() const {
  return _model;
}

const std::string& OccamDeviceBase::serial() const {
  return _serial;
}

OccamDeviceBase::ParamInfo* OccamDeviceBase::getParam(OccamDeviceParam id) {
  for (ParamInfo& pi : _params)
    if (pi.id == id)
      return &pi;
  return 0;
}

void OccamDeviceBase::registerParam(OccamDeviceParam id,
				    const std::string& name,
				    int min_value,
				    int max_value,
				    std::function<int()> geti_fn,
				    std::function<void(int)> seti_fn) {
  ParamInfo& pi = *_params.emplace(_params.end());
  pi.id = id;
  pi.name = name;
  pi.type = OCCAM_PARAM_INT;
  pi.min_value = double(min_value);
  pi.max_value = double(max_value);
  pi.count = 0;
  pi.geti_fn = geti_fn;
  pi.seti_fn = seti_fn;
}

void OccamDeviceBase::registerParam(OccamDeviceParam id,
				    const std::string& name,
				    double min_value,
				    double max_value,
				    std::function<double()> getr_fn,
				    std::function<void(double)> setr_fn) {
  ParamInfo& pi = *_params.emplace(_params.end());
  pi.id = id;
  pi.name = name;
  pi.type = OCCAM_PARAM_REAL;
  pi.min_value = min_value;
  pi.max_value = max_value;
  pi.count = 0;
  pi.getr_fn = getr_fn;
  pi.setr_fn = setr_fn;
}

void OccamDeviceBase::registerParam(OccamDeviceParam id,
				    const std::string& name,
				    std::function<std::string()> gets_fn,
				    std::function<void(const std::string&)> sets_fn) {
  ParamInfo& pi = *_params.emplace(_params.end());
  pi.id = id;
  pi.name = name;
  pi.type = OCCAM_PARAM_STRING;
  pi.min_value = 0;
  pi.max_value = 0;
  pi.count = 0;
  pi.gets_fn = gets_fn;
  pi.sets_fn = sets_fn;
}

void OccamDeviceBase::registerParam(OccamDeviceParam id,
				    const std::string& name,
				    int min_value,
				    int max_value,
				    int count,
				    std::function<int(int*,int)> getiv_fn,
				    std::function<int(const int*,int)> setiv_fn) {
  ParamInfo& pi = *_params.emplace(_params.end());
  pi.id = id;
  pi.name = name;
  pi.type = OCCAM_PARAM_INT;
  pi.min_value = double(min_value);
  pi.max_value = double(max_value);
  pi.count = count;
  pi.getiv_fn = getiv_fn;
  pi.setiv_fn = setiv_fn;
}

void OccamDeviceBase::registerParam(OccamDeviceParam id,
				    const std::string& name,
				    double min_value,
				    double max_value,
				    int count,
				    std::function<int(double*,int)> getrv_fn,
				    std::function<int(const double*,int)> setrv_fn) {
  ParamInfo& pi = *_params.emplace(_params.end());
  pi.id = id;
  pi.name = name;
  pi.type = OCCAM_PARAM_REAL;
  pi.min_value = min_value;
  pi.max_value = max_value;
  pi.count = count;
  pi.getrv_fn = getrv_fn;
  pi.setrv_fn = setrv_fn;
}

void OccamDeviceBase::registerParam(OccamDeviceParam id,
				    const std::string& name,
				    int count,
				    std::function<int(char**,int)> getsv_fn,
				    std::function<int(const char* const*,int)> setsv_fn) {
  ParamInfo& pi = *_params.emplace(_params.end());
  pi.id = id;
  pi.name = name;
  pi.type = OCCAM_PARAM_STRING;
  pi.min_value = 0;
  pi.max_value = 0;
  pi.count = count;
  pi.getsv_fn = getsv_fn;
  pi.setsv_fn = setsv_fn;
}

int OccamDeviceBase::enumerateDeviceParamList(OccamDeviceParamList** param_list) {
  *param_list = new OccamDeviceParamList;
  (*param_list)->params = new OccamDeviceParamEntry[_params.size()];
  (*param_list)->param_count = _params.size();
  for (int j=0;j<_params.size();++j) {
    ParamInfo& p0 = _params[j];
    OccamDeviceParamEntry& p1 = (*param_list)->params[j];
    memset(&p1,0,sizeof(OccamDeviceParamEntry));
    p1.id = p0.id;
    p1.name = strdup(p0.name.c_str());
    p1.type = p0.type;
    p1.min_value = p0.min_value;
    p1.max_value = p0.max_value;
    p1.count = p0.count;
  }
  return OCCAM_API_SUCCESS;
}

int OccamDeviceBase::setDeviceValuei(OccamDeviceParam id, int value) {
  ParamInfo* pi = getParam(id);
  if (!pi)
    return OCCAM_API_INVALID_PARAMETER;
  if (!bool(pi->seti_fn))
    return OCCAM_API_INVALID_TYPE;
  try {
    pi->seti_fn(value);
    return OCCAM_API_SUCCESS;
  } catch (RegisterProgramError& e) {
    return OCCAM_API_WRITE_ERROR;
  }
}

int OccamDeviceBase::setDeviceValuer(OccamDeviceParam id, double value) {
  ParamInfo* pi = getParam(id);
  if (!pi)
    return OCCAM_API_INVALID_PARAMETER;
  if (!bool(pi->setr_fn))
    return OCCAM_API_INVALID_TYPE;
  try {
    pi->setr_fn(value);
    return OCCAM_API_SUCCESS;
  } catch (RegisterProgramError& e) {
    return OCCAM_API_WRITE_ERROR;
  }
}

int OccamDeviceBase::setDeviceValues(OccamDeviceParam id, const char* value) {
  ParamInfo* pi = getParam(id);
  if (!pi)
    return OCCAM_API_INVALID_PARAMETER;
  if (!bool(pi->sets_fn))
    return OCCAM_API_INVALID_TYPE;
  try {
    pi->sets_fn(value);
    return OCCAM_API_SUCCESS;
  } catch (RegisterProgramError& e) {
    return OCCAM_API_WRITE_ERROR;
  }
}

int OccamDeviceBase::setDeviceValueiv(OccamDeviceParam id, const int* values, int value_count) {
  ParamInfo* pi = getParam(id);
  if (!pi)
    return OCCAM_API_INVALID_PARAMETER;
  if (!bool(pi->setiv_fn))
    return OCCAM_API_INVALID_TYPE;
  return pi->setiv_fn(values, value_count);
}

int OccamDeviceBase::setDeviceValuerv(OccamDeviceParam id, const double* values, int value_count) {
  ParamInfo* pi = getParam(id);
  if (!pi)
    return OCCAM_API_INVALID_PARAMETER;
  if (!bool(pi->setrv_fn))
    return OCCAM_API_INVALID_TYPE;
  return pi->setrv_fn(values, value_count);
}

int OccamDeviceBase::setDeviceValuesv(OccamDeviceParam id, const char* const* values, int value_count) {
  ParamInfo* pi = getParam(id);
  if (!pi)
    return OCCAM_API_INVALID_PARAMETER;
  if (!bool(pi->setsv_fn))
    return OCCAM_API_INVALID_TYPE;
  return pi->setsv_fn(values, value_count);
}

int OccamDeviceBase::getDeviceValuei(OccamDeviceParam id, int* value) {
  ParamInfo* pi = getParam(id);
  if (!pi)
    return OCCAM_API_INVALID_PARAMETER;
  if (!bool(pi->geti_fn))
    return OCCAM_API_INVALID_TYPE;
  try {
    *value = pi->geti_fn();
  } catch (RegisterProgramError& e) {
    return OCCAM_API_READ_ERROR;
  }
}

int OccamDeviceBase::getDeviceValuer(OccamDeviceParam id, double* value) {
  ParamInfo* pi = getParam(id);
  if (!pi)
    return OCCAM_API_INVALID_PARAMETER;
  if (!bool(pi->getr_fn))
    return OCCAM_API_INVALID_TYPE;
  try {
    *value = pi->getr_fn();
  } catch (RegisterProgramError& e) {
    return OCCAM_API_READ_ERROR;
  }
}

int OccamDeviceBase::getDeviceValues(OccamDeviceParam id, char** value) {
  ParamInfo* pi = getParam(id);
  if (!pi)
    return OCCAM_API_INVALID_PARAMETER;
  if (!bool(pi->gets_fn))
    return OCCAM_API_INVALID_TYPE;
  try {
    *value = strdup(pi->gets_fn().c_str());
  } catch (RegisterProgramError& e) {
    return OCCAM_API_READ_ERROR;
  }
}

int OccamDeviceBase::getDeviceValueiv(OccamDeviceParam id, int* values, int value_count) {
  ParamInfo* pi = getParam(id);
  if (!pi)
    return OCCAM_API_INVALID_PARAMETER;
  if (!bool(pi->getiv_fn))
    return OCCAM_API_INVALID_TYPE;
  return pi->getiv_fn(values, value_count);
}

int OccamDeviceBase::getDeviceValuerv(OccamDeviceParam id, double* values, int value_count) {
  ParamInfo* pi = getParam(id);
  if (!pi)
    return OCCAM_API_INVALID_PARAMETER;
  if (!bool(pi->getrv_fn))
    return OCCAM_API_INVALID_TYPE;
  return pi->getrv_fn(values, value_count);
}

int OccamDeviceBase::getDeviceValuesv(OccamDeviceParam id, char** values, int value_count) {
  ParamInfo* pi = getParam(id);
  if (!pi)
    return OCCAM_API_INVALID_PARAMETER;
  if (!bool(pi->getsv_fn))
    return OCCAM_API_INVALID_TYPE;
  return pi->getsv_fn(values, value_count);
}

int OccamDeviceBase::writeRegister(uint32_t addr, uint32_t value) {
  return OCCAM_API_NOT_SUPPORTED;
}

int OccamDeviceBase::writeRegister(uint32_t addr, uint32_t write_value, uint32_t mask) {
  uint32_t old_value;
  int r = readRegister(addr, &old_value);
  uint32_t new_value = (old_value & ~mask) | write_value;
  return writeRegister(addr, new_value);
}

int OccamDeviceBase::readRegister(uint32_t addr, uint32_t* value) {
  return OCCAM_API_NOT_SUPPORTED;
}

RegisterProgramResponse OccamDeviceBase::program
(const char* prog_str, uint32_t v0, uint32_t v1, uint32_t v2,
 uint32_t v3, uint32_t v4, uint32_t v5, uint32_t v6) {
  RegisterProgram prog(prog_str);
  prog.bind(0,v0);
  prog.bind(1,v1);
  prog.bind(2,v2);
  prog.bind(3,v3);
  prog.bind(4,v4);
  prog.bind(5,v5);
  prog.bind(6,v6);
  return prog.execute(*this);
}

int OccamDeviceBase::writeStorage(uint32_t addr, uint32_t len, const uint8_t* data) {
  return OCCAM_API_NOT_SUPPORTED;
}

int OccamDeviceBase::readStorage(uint32_t addr, uint32_t len, uint8_t* data) {
  return OCCAM_API_NOT_SUPPORTED;
}

int OccamDeviceBase::readImage(OccamImage** image, int block) {
  return OCCAM_API_NOT_SUPPORTED;
}

