#pragma once

#include "indigo.h"
#include <functional>
#include <vector>
#include <string>

class OccamDeviceBase;

class RegisterProgramError : public std::exception {
  std::string msg;
  int error;
public:
  RegisterProgramError(const std::string& msg, int error = OCCAM_API_GENERIC_ERROR);
  virtual ~RegisterProgramError() throw();
  virtual const char* what() const throw();
  int errorCode() const;
};

class RegisterProgramResponse {
public:
  struct ResultInfo {
    uint32_t reg_index;
    uint32_t reg_value;
  };
private:
  std::vector<ResultInfo> results;
public:
  void add(uint32_t reg_index, uint32_t reg_value);
  void assign(int i, uint32_t reg_index, uint32_t reg_value);
  void clear();
  int size() const;
  uint32_t regIndex(int i) const;
  uint32_t regValue(int i) const;
  uint32_t operator[] (int i) const;
  std::string toString() const;
};

class RegisterProgram {
public:
  enum CommandType {
    CommandType_Read,
    CommandType_Write,
    CommandType_ReadWrite,
    CommandType_Sleep
  };
  struct CommandInfo {
    CommandType type;
    uint32_t reg_index;
    uint32_t reg_mask;
    uint32_t reg_value;
    uint64_t delay_us;
    bool late_bound;
  };
private:
  std::vector<CommandInfo> commands;
  template <class T>
  const char* parseNumber(const char* cmd_s, T& number, bool* late_bound = 0);
  const char* parseNextChar(const char* cmd_s, char& c, const char* expect_hint);
  const char* parseDelim(const char* cmd_s, char delim, bool optional);
  const char* parseCommandType(const char* cmd_s, CommandType& type);
  const char* parseCommand(const char* cmd_s, CommandInfo& cmd);
public:
  RegisterProgram();
  RegisterProgram(const std::string& reg_cmd);
  void add(const std::string& reg_cmd);
  void addRead(uint32_t reg_index);
  void addWrite(uint32_t reg_index, uint32_t reg_value);
  void addSleep(uint64_t delay_us);
  void bind(int index, uint32_t value);
  RegisterProgramResponse execute(OccamDeviceBase& dev);
  int size() const;
  const CommandInfo& operator[] (int index) const;
  CommandInfo& operator[] (int index);
};

class OccamDeviceBase {
  struct ParamInfo {
    OccamDeviceParam id;
    std::string name;
    OccamParamType type;
    double min_value;
    double max_value;
    int count;
    std::function<int()> geti_fn;
    std::function<void(int)> seti_fn;
    std::function<double()> getr_fn;
    std::function<void(double)> setr_fn;
    std::function<std::string()> gets_fn;
    std::function<void(const std::string&)> sets_fn;
    std::function<int(int*,int)> getiv_fn;
    std::function<int(const int*,int)> setiv_fn;
    std::function<int(double*,int)> getrv_fn;
    std::function<int(const double*,int)> setrv_fn;
    std::function<int(char**,int)> getsv_fn;
    std::function<int(const char* const*,int)> setsv_fn;
  };
  std::vector<ParamInfo> _params;
  std::string _cid;
  std::string _model;
  std::string _serial;
  ParamInfo* getParam(OccamDeviceParam id);
public:
  OccamDeviceBase(const std::string& cid);
  const std::string& cid() const;
  const std::string& model() const;
  const std::string& serial() const;
  virtual ~OccamDeviceBase();
  void registerParam(OccamDeviceParam id,
		     const std::string& name,
		     int min_value,
		     int max_value,
		     std::function<int()> geti_fn,
		     std::function<void(int)> seti_fn = std::function<void(int)>());
  void registerParam(OccamDeviceParam id,
		     const std::string& name,
		     double min_value,
		     double max_value,
		     std::function<double()> getr_fn,
		     std::function<void(double)> setr_fn = std::function<void(double)>());
  void registerParam(OccamDeviceParam id,
		     const std::string& name,
		     std::function<std::string()> gets_fn,
		     std::function<void(const std::string&)> sets_fn = std::function<void(const std::string&)>());
  void registerParam(OccamDeviceParam id,
		     const std::string& name,
		     int min_value,
		     int max_value,
		     int count,
		     std::function<int(int*,int)> getiv_fn,
		     std::function<int(const int*,int)> setiv_fn = std::function<int(const int*,int)>());
  void registerParam(OccamDeviceParam id,
		     const std::string& name,
		     double min_value,
		     double max_value,
		     int count,
		     std::function<int(double*,int)> getrv_fn,
		     std::function<int(const double*,int)> setrv_fn = std::function<int(const double*,int)>());
  void registerParam(OccamDeviceParam id,
		     const std::string& name,
		     int count,
		     std::function<int(char**,int)> getsv_fn,
		     std::function<int(const char* const*,int)> setsv_fn = std::function<int(const char* const*,int)>());

  int enumerateDeviceParamList(OccamDeviceParamList** param_list);

  int setDeviceValuei(OccamDeviceParam id, int value);
  int setDeviceValuer(OccamDeviceParam id, double value);
  int setDeviceValues(OccamDeviceParam id, const char* value);
  int setDeviceValueiv(OccamDeviceParam id, const int* values, int values_count);
  int setDeviceValuerv(OccamDeviceParam id, const double* values, int values_count);
  int setDeviceValuesv(OccamDeviceParam id, const char* const* values, int values_count);
  int getDeviceValuei(OccamDeviceParam id, int* value);
  int getDeviceValuer(OccamDeviceParam id, double* value);
  int getDeviceValues(OccamDeviceParam id, char** value);
  int getDeviceValueiv(OccamDeviceParam id, int* values, int value_count);
  int getDeviceValuerv(OccamDeviceParam id, double* value, int value_count);
  int getDeviceValuesv(OccamDeviceParam id, char** value, int value_count);

  virtual int writeRegister(uint32_t addr, uint32_t value);
  virtual int writeRegister(uint32_t addr, uint32_t value, uint32_t mask);
  virtual int readRegister(uint32_t addr, uint32_t* value);
  RegisterProgramResponse program(const char* prog_str, uint32_t v0=0, uint32_t v1=0, uint32_t v2=0,
				  uint32_t v3=0, uint32_t v4=0, uint32_t v5=0, uint32_t v6=0);

  virtual int writeStorage(uint32_t addr, uint32_t len, const uint8_t* data);
  virtual int readStorage(uint32_t addr, uint32_t len, uint8_t* data);

  virtual int readImage(OccamImage** image, int block);
};

void registerDevice(const std::string& model,
		    std::function<OccamDeviceBase*(const std::string&)> ctor_fn);
template <class T>
void registerDevice(const std::string& model) {
  registerDevice(model, [](const std::string& cid){
      return static_cast<OccamDeviceBase*>(new T(cid));
    });
}

// Local Variables:
// mode: c++
// End:
