#include "indigo.h"
#include "omni_libusb.h"
#include <assert.h>
#include <iostream>
#include <vector>
#include <string>
#include <sstream>
#include <list>
#include <memory>
#include <string.h>

//#define DEBUG_DATA_RATES

//////////////////////////////////////////////////////////////////////////////////
// OmniBuffer

void OmniBuffer::__transfer_done(struct libusb_transfer* xfer0) {
  ((OmniBuffer*)xfer0->user_data)->transfer_done(xfer0);
}

void OmniBuffer::transfer_done(struct libusb_transfer* xfer0) {
  assert(xfer == xfer0);
  pending = false;

  const char* status_str = "??";
  switch (xfer->status) {
  case LIBUSB_TRANSFER_COMPLETED: status_str = "COMPLETED"; break;
  case LIBUSB_TRANSFER_ERROR: status_str = "ERROR"; break;
  case LIBUSB_TRANSFER_TIMED_OUT: status_str = "TIMED_OUT"; break;
  case LIBUSB_TRANSFER_CANCELLED: status_str = "CANCELLED"; break;
  case LIBUSB_TRANSFER_STALL: status_str = "STALL"; break;
  case LIBUSB_TRANSFER_NO_DEVICE: status_str = "NO_DEVICE"; break;
  case LIBUSB_TRANSFER_OVERFLOW: status_str = "OVERFLOW"; break;
  }

  if (xfer->status != LIBUSB_TRANSFER_COMPLETED) {
    std::cerr<<"transfer "<<xfer<<" done, status = "<<status_str<<std::endl;
    return;
  }
}

OmniBuffer::OmniBuffer(libusb_device_handle* handle,
		       int size,
		       int endpoint)
  : xfer(0), pending(false) {
  _data.resize(size);

  xfer = libusb_alloc_transfer(0);
  if (!xfer)
    return;
  libusb_fill_bulk_transfer(xfer, handle, endpoint,
			    &_data[0], _data.size(),
			    __transfer_done, this, 2000);
}

OmniBuffer::~OmniBuffer() {
  // these will be freed synchronously when the device handle is closed
  // if (xfer) {
  //   if (pending)
  //     libusb_cancel_transfer(xfer);
  //   libusb_free_transfer(xfer);
  // }
}

void OmniBuffer::submit() {
  assert(!pending);
  pending = true;
  int r = libusb_submit_transfer(xfer);
}

bool OmniBuffer::isPending() const {
  return pending;
}

const uint8_t* OmniBuffer::data() const {
  return &_data[0];
}

uint8_t* OmniBuffer::data() {
  return &_data[0];
}

int OmniBuffer::size() const {
  return _data.size();
}

//////////////////////////////////////////////////////////////////////////////////
// OmniParseFIFO

int OmniParseFIFO::remaining() {
  int remaining8 = 0;
  for (BufferInfo& bi : bufs)
    remaining8 += bi.remaining8;
  return remaining8;
}

void OmniParseFIFO::consume(void* dst, int n) {
  uint8_t* dst8 = (uint8_t*)dst;
  for (BufferInfo& bi : bufs) {
    int cplen = (std::min)(bi.remaining8,n);
    std::copy(bi.buf->data()+bi.pos8,
	      bi.buf->data()+bi.pos8+cplen,
	      (uint8_t*)dst8);
    bi.pos8 += cplen;
    bi.remaining8 -= cplen;
    dst8 += cplen;
    n -= cplen;
    if (!n)
      return;
  }
  assert(0);
}

uint32_t OmniParseFIFO::consume32() {
  uint32_t v;
  consume(&v, sizeof(uint32_t));
  return v;
}

void OmniParseFIFO::push(std::shared_ptr<OmniBuffer> buf) {
  if (buf->size()&3) {
    assert(0);
    return;
  }
  BufferInfo& bi = *bufs.insert(bufs.end(),BufferInfo());
  bi.buf = buf;
  bi.pos8 = 0;
  bi.remaining8 = buf->size();
}

void OmniParseFIFO::clear() {
  for (BufferInfo& bi : bufs)
    bi.remaining8 = 0;
}

void OmniParseFIFO::flush(std::list<std::shared_ptr<OmniBuffer> >& buffers) {
  while (!bufs.empty() && !bufs.begin()->remaining8) {
    bufs.begin()->buf->submit();
    buffers.push_back(bufs.begin()->buf);
    bufs.erase(bufs.begin());
  }
}

//////////////////////////////////////////////////////////////////////////////////
// OmniDevice

void OmniDevice::newDataAvailable(std::shared_ptr<OmniBuffer> buf) {
#ifdef DEBUG_DATA_RATES
  if (xfer_rate.increment((*buffers.begin())->size())) {
    std::cerr<<"fx3 xfer rate: "<<xfer_rate.rateMBs()<<" MB/s"<<std::endl;
  }
#endif // DEBUG_DATA_RATES

  if (!image_fr)
    initFrame();

  stream.push(buf);
  stream.flush(buffers);

  bool underrun = false;
  while (stream.remaining()>0&&!underrun) {
    switch (state) {
    case 0: {
      if (stream.consume32() == 0x4819bd4b)
	state = 1;
      break;
    }
    case 1: {
      addr16 = stream.consume32();
      addr16_noflags = (addr16&((1<<22)-1));
      addr16_noflags %= 5*752*480/2;

      uint32_t addr8 = addr16_noflags*2;

      sensor_index = addr8 / (752*480);
      int page_index = addr8 / 512;
      if (page_index == 0)
	state = 0;
      else {
	state = 2;
	int new_pixel_addr8 = page_index * 512;
	if (new_pixel_addr8 < pixel_addr8) {
	  const int hz = 60;
	  image_fr->index = frame_count++;
	  image_fr->time_ns = uint64_t(image_fr->index) * uint64_t(1000 * 1000 * 1000) / uint64_t(hz);

	  emitFrame(image_fr);
	  initFrame();
	}
	pixel_addr8 = new_pixel_addr8;
      }
      break;
    }
    case 2: {
      if (pixel_addr8<0 || pixel_addr8>=pixels_size) {
	state = 0;
	break;
      }
      if (stream.remaining()<512) {
	underrun = true;
	break;
      }
      stream.consume(image_fr->data[0]+pixel_addr8,512);

      state = 0;
      break;
    }
    default: {
      state = 0;
      break;
    }
    }
  }
}

void OmniDevice::initFrame() {
  image_fr = new OccamImage;
  image_fr->cid = strdup(cid.c_str());
  memset(image_fr->timescale,0,sizeof(image_fr->timescale));
  image_fr->refcnt = 1;
  image_fr->backend = OCCAM_LOCAL;
  image_fr->format = OCCAM_GRAY8;
  image_fr->width = 752;
  image_fr->height = 480*5;
  memset(image_fr->step,0,sizeof(image_fr->step));
  memset(image_fr->data,0,sizeof(image_fr->data));
  image_fr->step[0] = 752;
  image_fr->data[0] = new uint8_t[image_fr->width*image_fr->height];
}

void OmniDevice::emitFrame(OccamImage* image_fr) {
#ifdef DEBUG_DATA_RATES
  if (frame_rate.increment()) {
    std::cerr<<"FPS: "<<frame_rate.rate()<<std::endl;
  }
#endif // DEBUG_DATA_RATES

  queued_frames.push_back(image_fr);
  ++queued_frames_count;
  while (queued_frames_count > max_queued_frames) {
    --queued_frames_count;
    assert(!queued_frames.empty());
    occamFreeImage(*queued_frames.begin());
    queued_frames.pop_front();
  }
}

OmniDevice::OmniDevice(const std::string& cid,
		       int _max_buffers)
  : OccamDeviceBase(cid),
    handle(0),
    buffer_count(0),
    max_buffers(_max_buffers),
    image_fr(0),
    pixels_pos(0),
    pixels_size(752*480*5),
    frame_count(0),
    max_queued_frames(1),
    queued_frames_count(0),
    state(0) {

  auto get_max_queued_frames = [this]() {
    return this->max_queued_frames;
  };
  auto set_max_queued_frames = [this](int value) {
    this->max_queued_frames = value;
  };
  registerParam(OCCAM_MAX_USB_QUEUED_FRAMES, "max_usb_queued_frames", 1, 32,
		get_max_queued_frames, set_max_queued_frames);

  libusb_device** device_list;
  ssize_t num_devices = libusb_get_device_list(0, &device_list);

  for (int j=0;j<num_devices;++j) {
    libusb_device* dev = device_list[j];
    struct libusb_device_descriptor desc;
    int r = libusb_get_device_descriptor(dev,&desc);

    if (desc.idVendor != 0x285e)
      continue;

    libusb_device_handle* handlej;
    r = libusb_open(dev, &handlej);

    char model[1024];
    char serial[1024];
    r = libusb_get_string_descriptor_ascii(handlej, desc.iProduct, (unsigned char*)model, sizeof(model));
    r = libusb_get_string_descriptor_ascii(handlej, desc.iSerialNumber, (unsigned char*)serial, sizeof(serial));

    std::stringstream cidj;
    cidj<<model<<":"<<serial;
    if (cidj.str() == cid) {
      handle = handlej;
      break;
    }

    libusb_close(handlej);
  }

  libusb_free_device_list(device_list,1);
}

OmniDevice::~OmniDevice() {
  buffers.clear();

  for (OccamImage* img : queued_frames)
    occamFreeImage(img);

  if (handle)
    libusb_close(handle);
}

bool OmniDevice::isOpen() const {
  return !!handle;
}

int OmniDevice::writeRegister(uint32_t addr, uint32_t value) {
  uint64_t index_value;
  uint32_t* index_value32 = (uint32_t*)&index_value;
  index_value32[0] = addr;
  index_value32[1] = value;

  int r = libusb_control_transfer
    (handle, LIBUSB_RECIPIENT_DEVICE|LIBUSB_REQUEST_TYPE_VENDOR|LIBUSB_ENDPOINT_OUT,
     0xc0, 0, 0, (unsigned char*)&index_value, sizeof(index_value), 1000);
  if (r != 8)
    return OCCAM_API_WRITE_ERROR;

  return OCCAM_API_SUCCESS;
}

int OmniDevice::readRegister(uint32_t addr, uint32_t* value) {
  int r;
  if ((r = writeRegister(0xcc00, addr)) != OCCAM_API_SUCCESS)
    return r;

  r = libusb_control_transfer
    (handle, LIBUSB_RECIPIENT_DEVICE|LIBUSB_REQUEST_TYPE_VENDOR|LIBUSB_ENDPOINT_IN,
     0xc1, 0, 0, (unsigned char*)value, sizeof(value), 1000);
  if (r != 4)
    return OCCAM_API_WRITE_ERROR;

  return OCCAM_API_SUCCESS;
}

int OmniDevice::writeStorage(uint32_t addr, uint32_t len, const uint8_t* data) {
  return OCCAM_API_NOT_SUPPORTED;
}

int OmniDevice::readStorage(uint32_t addr, uint32_t len, uint8_t* data) {
  return OCCAM_API_NOT_SUPPORTED;
}

int OmniDevice::readImage(OccamImage** image, int block) {
  struct timeval tv;
  tv.tv_sec = 0;
  tv.tv_usec = 0;
  int completed = 0;
  int r = libusb_handle_events_timeout_completed(0, &tv, &completed);
  if (r)
    return OCCAM_API_READ_ERROR;

  for (;;) {
    update();

    if (!queued_frames.empty()) {
      *image = *queued_frames.begin();
      queued_frames.pop_front();
      assert(queued_frames_count>0);
      --queued_frames_count;
      return OCCAM_API_SUCCESS;
    }

    if (!block)
      return OCCAM_API_IMAGE_NOT_AVAILABLE;

    struct timeval tv;
    tv.tv_sec = 0;
    tv.tv_usec = 1;
    int completed = 0;
    int r = libusb_handle_events_timeout_completed(0, &tv, &completed);
    if (r)
      return OCCAM_API_READ_ERROR;
  }
}

void OmniDevice::update() {
  while (buffer_count < max_buffers) {
    auto b0 = std::make_shared<OmniBuffer>(handle);
    buffers.push_back(b0);
    ++buffer_count;
    b0->submit();
  }
  if (buffers.empty())
    return;
  while (!(*buffers.begin())->isPending()) {
    std::shared_ptr<OmniBuffer> buf = *buffers.begin();
    buffers.pop_front();
    newDataAvailable(buf);
  }
}
