#include "simple_cpu_blend.h"
#include <string.h>
#include <assert.h>
#include <algorithm>

OccamImage* simpleCPUBlend(const OccamImage* mfr0,
			   int N,
			   const int* si_x,
			   const int* si_y,
			   const int* si_width,
			   const int* si_height,
			   const int* offsetx,
			   const int* offsety) {
  assert(mfr0->backend == OCCAM_LOCAL);
  assert(mfr0->format == OCCAM_GRAY8 || mfr0->format == OCCAM_RGB24);

  OccamImage* fr1 = new OccamImage;
  fr1->cid = strdup(mfr0->cid);
  memcpy(fr1->timescale,mfr0->timescale,sizeof(fr1->timescale));
  fr1->time_ns = mfr0->time_ns;
  fr1->index = mfr0->index;
  fr1->refcnt = 1;
  fr1->backend = mfr0->backend;
  fr1->format = mfr0->format;
  int channels = 1;
  switch (fr1->format) {
  case OCCAM_GRAY8: channels = 1; break;
  case OCCAM_RGB24: channels = 3; break;
  default:
    assert(0);
    return 0;
  }

  // create output image and calculate dims
  int width = 0;
  int height = 0;
  for (int si_j=0;si_j<N;++si_j) {
    width += si_width[si_j];
    height = std::max(height,si_height[si_j]);
  }
  fr1->width = width;
  fr1->height = height;
  memset(fr1->step,0,sizeof(fr1->step));
  memset(fr1->data,0,sizeof(fr1->data));
  fr1->step[0] = (width+15)&~15;
  fr1->data[0] = new uint8_t[fr1->height*fr1->step[0]];

  int left_x = width;
  int right_x = 0;
  int top_y = 0;
  int bottom_y = height;

  // copy the images
  std::vector<int> dstx0;
  dstx0.reserve(N);
  int dstx = 0;
  int n = 0;
  for (int si_j=0;si_j<N;++si_j) {
    int linebytes = si_width[si_j]*channels;
    int dst_step = fr1->step[0];
    int src_step = mfr0->step[0];
    uint8_t* dstp = fr1->data[0]+dstx*channels;
    const uint8_t* srcp = mfr0->data[0]+src_step*si_y[si_j]+si_x[si_j]*channels;

    int offy = std::max(-si_height[si_j],std::min(si_height[si_j],n>=N?0:offsety[n]));
    int si_heightj = si_height[si_j];
    if (offy<0) {
      si_heightj += offy;
      srcp += src_step*-offy;
    } else {
      si_heightj -= offy;
      dstp += dst_step*offy;
    }

    for (int y=0;y<si_heightj;++y,srcp+=src_step,dstp+=dst_step)
      memcpy(dstp,srcp,linebytes);

    left_x = std::max(0,std::min(left_x,dstx));
    right_x = std::min(width,std::max(right_x,dstx+si_width[si_j]));
    top_y = std::max(0,std::max(top_y,offy));
    bottom_y = std::min(std::min(bottom_y,si_height[si_j]+offy),height);

    dstx0.push_back(dstx);
    dstx += si_width[si_j];
    dstx -= std::min(si_width[si_j],n>=N?0:offsetx[n]);
    ++n;
  }

  // blend overlap regions
  for (int j=0;j<N-1;++j) {
    int si0 = j, si1 = j+1;
    int img0_dst_x = dstx0[j];
    int img1_dst_x = dstx0[j+1];
    int overlap_width = img0_dst_x + si_width[si0] - img1_dst_x;
    if (!overlap_width)
      continue;

    int offy0 = j>=N?0:std::max(-si_height[si0],std::min(si_height[si0],offsety[j]));
    int offy1 = j+1>=N?0:std::max(-si_height[si1],std::min(si_height[si1],offsety[j+1]));

    int dst_step = fr1->step[0];
    int src_step = mfr0->step[0];
    uint8_t* dstp = fr1->data[0]+img1_dst_x*channels;
    const uint8_t* srcp = mfr0->data[0];
    int fade_step = (2<<16) / overlap_width;
    for (int y=0;y<height;++y,dstp+=dst_step) {
      int y0 = y-offy0;
      if (y0 < 0 || y0 >= si_height[si0])
	continue;
      int y1 = y-offy1;
      if (y1 < 0 || y1 >= si_height[si1])
	continue;
      const uint8_t* srcp0 = srcp+src_step*(si_y[si0]+y0)+(si_x[si0]+si_width[si0]-overlap_width)*channels;
      const uint8_t* srcp1 = srcp+src_step*(si_y[si1]+y1)+si_x[si1]*channels;

      int fade0 = (2<<16), fade1 = 0;
      for (int x=0,xx=overlap_width*channels;x<xx;x+=channels) {
	for (int c=0;c<channels;++c) {
	  int v0 = int(srcp0[x+c]);
	  int v1 = int(srcp1[x+c]);
	  int v = ((v0*fade0)>>17) + ((v1*fade1)>>17);
	  dstp[x+c] = v;
	}
	fade0 -= fade_step;
	fade1 += fade_step;
      }
    }
  }

  // crop the output image such that every pixel is filled by some input image
  OccamImage* fr2 = new OccamImage;
  fr2->cid = strdup(mfr0->cid);
  memcpy(fr2->timescale,mfr0->timescale,sizeof(fr1->timescale));
  fr2->time_ns = mfr0->time_ns;
  fr2->index = mfr0->index;
  fr2->refcnt = 1;
  fr2->backend = mfr0->backend;
  fr2->format = mfr0->format;
  fr2->width = right_x-left_x;
  fr2->height = bottom_y-top_y;
  memset(fr2->step,0,sizeof(fr2->step));
  memset(fr2->data,0,sizeof(fr2->data));
  fr2->step[0] = (fr2->width+15)&~15;
  fr2->data[0] = new uint8_t[fr2->height*fr2->step[0]];
  int linebytes = fr2->width*channels;
  int src_step = fr1->step[0];
  int dst_step = fr2->step[0];
  const uint8_t* srcp = fr1->data[0]+top_y*src_step+left_x*channels;
  uint8_t* dstp = fr2->data[0];
  for (int y=0,yy=fr2->height;y<yy;++y,srcp+=src_step,dstp+=dst_step)
    memcpy(dstp,srcp,linebytes);

  occamFreeImage(fr1);
  return fr2;
}
