#include <thread>
#include <mutex>
#include <condition_variable>

template <class T>
struct ExternalDestroyTraits {
  static void destroy(T value) {
  }
};

template <>
struct ExternalDestroyTraits<OccamImage*> {
  static void destroy(OccamImage* value) {
    occamFreeImage(value);
  }
};

template <class R, class T>
class ThreadedFilter {
  std::thread th;
  bool shutdown;
  mutable std::mutex mut;
  std::condition_variable cond;
  std::function<R(const T)> proc;
  std::list<T> pending;
  std::list<R> reaping;
  int pending_count;
  int max_pending;

  void threadproc() {
    while (1) {
      T in_value;
      {
	std::unique_lock<std::mutex> g(mut);
	if (shutdown)
	  break;
	if (pending.empty()) {
	  cond.wait(g);
	  continue;
	}
	in_value = *pending.begin();
	pending.pop_front();
	--pending_count;
      }

      R out_value = proc(in_value);

      {
	std::unique_lock<std::mutex> g(mut);
	ExternalDestroyTraits<T>::destroy(in_value);
	reaping.push_back(out_value);
      }
    }
  }
public:
  ThreadedFilter(std::function<R(const T)> _proc, int _max_pending = 2)
    : proc(_proc),
      shutdown(false),
      pending_count(0),
      max_pending(_max_pending) {
    th = std::thread([this](){
	this->threadproc();
      });
  }
  ~ThreadedFilter() {
    {
      std::unique_lock<std::mutex> g(mut);
      shutdown = true;
      cond.notify_all();
    }
    th.join();
  }
  ThreadedFilter(const ThreadedFilter& x) = delete;
  ThreadedFilter& operator= (const ThreadedFilter& rhs) = delete;

  void setMaxPending(int _max_pending) {
    std::unique_lock<std::mutex> g(mut);
    max_pending = _max_pending;
  }
  int maxPending() const {
    std::unique_lock<std::mutex> g(mut);
    return max_pending;
  }

  void push(const T& value) {
    std::unique_lock<std::mutex> g(mut);
    pending.push_back(value);
    pending_count++;
    while (pending_count>max_pending) {
      ExternalDestroyTraits<T>::destroy(*pending.begin());
      pending.pop_front();
      pending_count--;
    }
    cond.notify_one();
  }

  bool pop(R& value) {
    std::unique_lock<std::mutex> g(mut);
    if (reaping.empty())
      return false;
    value = *reaping.begin();
    reaping.pop_front();
    return true;
  }
};

typedef ThreadedFilter<OccamImage*,OccamImage*> ThreadedImageFilter;

// Local Variables:
// mode: c++
// End:
